@@ -3,7 +3,8 @@
  */
 
 /*
- * Copyright (C) 2012-2019, 2021, the Free Software Foundation, Inc.
+ * Copyright (C) 2012-2019, 2021, 2022, 2023, 2024,
+ * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -69,19 +70,20 @@ api_get_argument(awk_ext_id_t id, size_t count,
 		return awk_false;
 
 	/* if type is undefined */
-	if (arg->type == Node_var_new) {
+	if (arg->type == Node_var_new || arg->type == Node_elem_new) {
 		if (wanted == AWK_UNDEFINED)
 			return awk_true;
-		else if (wanted == AWK_ARRAY) {
+		else if (wanted == AWK_ARRAY)
 			goto array;
-		} else {
+		else
 			goto scalar;
-		}
 	}
 
 	/* at this point, we have real type */
 	if (arg->type == Node_var_array || arg->type == Node_array_ref) {
-		if (wanted != AWK_ARRAY && wanted != AWK_UNDEFINED)
+		if (wanted == AWK_UNDEFINED)
+			return awk_true;
+		else if (wanted != AWK_ARRAY)
 			return awk_false;
 		goto array;
 	} else
@@ -124,7 +126,7 @@ api_set_argument(awk_ext_id_t id,
 		return awk_false;
 
 	if (   (arg = get_argument(count)) == NULL
-	    || arg->type != Node_var_new)
+	    || (arg->type != Node_var_new && arg->type != Node_elem_new))
 		return awk_false;
 
 	arg = get_array_argument(arg, count);
@@ -148,7 +150,9 @@ awk_value_to_node(const awk_value_t *retval)
 {
 	NODE *ext_ret_val = NULL;
 	NODE *v;
+#ifdef HAVE_MPFR
 	int tval = 0;
+#endif
 
 	if (retval == NULL)
 		fatal(_("awk_value_to_node: received null retval"));
@@ -160,6 +164,9 @@ awk_value_to_node(const awk_value_t *retval)
 	case AWK_UNDEFINED:
 		ext_ret_val = dupnode(Nnull_string);
 		break;
+	case AWK_BOOL:
+		ext_ret_val = make_bool_node(retval->bool_value != awk_false);
+		break;
 	case AWK_NUMBER:
 		switch (retval->num_type) {
 		case AWK_NUMBER_TYPE_DOUBLE:
@@ -170,8 +177,10 @@ awk_value_to_node(const awk_value_t *retval)
 			if (! do_mpfr)
 				fatal(_("awk_value_to_node: not in MPFR mode"));
 			ext_ret_val = make_number_node(MPFN);
-			tval = mpfr_set(ext_ret_val->mpg_numbr, (mpfr_ptr) retval->num_ptr, ROUND_MODE);
+			mpfr_init(ext_ret_val->mpg_numbr);
+			tval = mpfr_set(ext_ret_val->mpg_numbr, (mpfr_srcptr) retval->num_ptr, ROUND_MODE);
 			IEEE_FMT(ext_ret_val->mpg_numbr, tval);
+			mpfr_clear(retval->num_ptr);
 #else
 			fatal(_("awk_value_to_node: MPFR not supported"));
 #endif
@@ -181,7 +190,9 @@ awk_value_to_node(const awk_value_t *retval)
 			if (! do_mpfr)
 				fatal(_("awk_value_to_node: not in MPFR mode"));
 			ext_ret_val = make_number_node(MPZN);
+			mpz_init(ext_ret_val->mpg_i);
 			mpz_set(ext_ret_val->mpg_i, (mpz_ptr) retval->num_ptr);
+			mpz_clear(retval->num_ptr);
 #else
 			fatal(_("awk_value_to_node: MPFR not supported"));
 #endif
@@ -532,6 +543,16 @@ assign_regex(NODE *node, awk_value_t *val)
 	val->val_type = AWK_REGEX;
 }
 
+/* assign_bool --- return a bool node */
+
+static inline void
+assign_bool(NODE *node, awk_value_t *val)
+{
+	assert((node->flags & BOOLVAL) != 0);
+	val->val_type = AWK_BOOL;
+	val->bool_value = get_number_si(node) != 0 ? awk_true : awk_false;
+}
+
 /* node_to_awk_value --- convert a node into a value for an extension */
 
 static awk_bool_t
@@ -547,6 +568,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 
 	switch (node->type) {
 	case Node_var_new:	/* undefined variable */
+	case Node_elem_new:	/* undefined element */
 		val->val_type = AWK_UNDEFINED;
 		if (wanted == AWK_UNDEFINED) {
 			ret = awk_true;
@@ -567,8 +589,16 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 	case Node_val:
 		/* a scalar value */
 		switch (wanted) {
+		case AWK_BOOL:
+			if ((node->flags & BOOLVAL) != 0) {
+				assign_bool(node, val);
+				ret = awk_true;
+			} else
+				ret = awk_false;
+			break;
+
 		case AWK_NUMBER:
-			if (node->flags & REGEX)
+			if ((node->flags & REGEX) != 0)
 				val->val_type = AWK_REGEX;
 			else {
 				(void) force_number(node);
@@ -578,7 +608,10 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 			break;
 
 		case AWK_STRNUM:
-			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX)) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
 			case STRING:
 				val->val_type = AWK_STRING;
 				break;
@@ -593,11 +626,9 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				val->val_type = AWK_REGEX;
 				break;
 			case NUMBER|STRING:
-				if (node == Nnull_string) {
-					val->val_type = AWK_UNDEFINED;
-					break;
-				}
-				/* fall through */
+				// this can come from a Node_elem_new, as well as Nnull_string
+				val->val_type = AWK_UNDEFINED;
+				break;
 			default:
 				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
@@ -612,10 +643,13 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 			break;
 
 		case AWK_REGEX:
-			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX)) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
 			case STRING:
 				val->val_type = AWK_STRING;
 				break;
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
 			case NUMBER:
 				val->val_type = AWK_NUMBER;
 				break;
@@ -627,11 +661,9 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				ret = awk_true;
 				break;
 			case NUMBER|STRING:
-				if (node == Nnull_string) {
-					val->val_type = AWK_UNDEFINED;
-					break;
-				}
-				/* fall through */
+				// this can come from a Node_elem_new, as well as Nnull_string
+				val->val_type = AWK_UNDEFINED;
+				break;
 			default:
 				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
@@ -640,7 +672,10 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 			break;
 
 		case AWK_SCALAR:
-			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX)) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				val->val_type = AWK_BOOL;
+				break;
 			case STRING:
 				val->val_type = AWK_STRING;
 				break;
@@ -654,11 +689,9 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				val->val_type = AWK_REGEX;
 				break;
 			case NUMBER|STRING:
-				if (node == Nnull_string) {
-					val->val_type = AWK_UNDEFINED;
-					break;
-				}
-				/* fall through */
+				// this can come from a Node_elem_new, as well as Nnull_string
+				val->val_type = AWK_UNDEFINED;
+				break;
 			default:
 				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
@@ -668,7 +701,11 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 
 		case AWK_UNDEFINED:
 			/* return true and actual type for request of undefined */
-			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX)) {
+			switch (fixtype(node)->flags & (STRING|NUMBER|USER_INPUT|REGEX|BOOLVAL)) {
+			case NUMBER|BOOLVAL:
+				assign_bool(node, val);
+				ret = awk_true;
+				break;
 			case STRING:
 				assign_string(node, val, AWK_STRING);
 				ret = awk_true;
@@ -686,12 +723,10 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				ret = awk_true;
 				break;
 			case NUMBER|STRING:
-				if (node == Nnull_string) {
-					val->val_type = AWK_UNDEFINED;
-					ret = awk_true;
-					break;
-				}
-				/* fall through */
+				// this can come from a Node_elem_new, as well as Nnull_string
+				val->val_type = AWK_UNDEFINED;
+				ret = awk_true;
+				break;
 			default:
 				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
@@ -846,7 +881,8 @@ api_sym_update(awk_ext_id_t id,
 
 	/*
 	 * If we get here, then it exists already.  Any valid type is
-	 * OK except for AWK_ARRAY.
+	 * OK except for AWK_ARRAY (unless it is in Node_var_new undefined
+	 * state, in which case an array is OK).
 	 */
 	if (   (node->flags & NO_EXT_SET) != 0
 	    || is_off_limits_var(full_name)) {	/* most built-in vars not allowed */
@@ -857,11 +893,24 @@ api_sym_update(awk_ext_id_t id,
 
 	efree((void *) full_name);
 
-	if (    value->val_type != AWK_ARRAY
-	    && (node->type == Node_var || node->type == Node_var_new)) {
+	if (value->val_type == AWK_ARRAY) {
+		if (node->type == Node_var_new) {
+			/* special gymnastics to convert untyped to an array */
+			array_node = awk_value_to_node(value);
+			array_node->vname = node->vname;
+			unref(node->var_value);
+			*node = *array_node;
+			freenode(array_node);
+			value->array_cookie = node;	/* pass new cookie back to extension */
+			return awk_true;
+		}
+	} else if (node->type == Node_var
+		   || node->type == Node_var_new
+		   || node->type == Node_elem_new) {
 		unref(node->var_value);
 		node->var_value = awk_value_to_node(value);
-		if (node->type == Node_var_new && value->val_type != AWK_UNDEFINED)
+		if ((node->type == Node_var_new || node->type == Node_elem_new)
+		    && value->val_type != AWK_UNDEFINED)
 			node->type = Node_var;
 
 		return awk_true;
@@ -1166,6 +1215,17 @@ api_clear_array(awk_ext_id_t id, awk_array_t a_cookie)
 	return awk_true;
 }
 
+/* api_destroy_array --- destroy an array */
+
+static awk_bool_t
+api_destroy_array(awk_ext_id_t id, awk_array_t a_cookie)
+{
+	if (! api_clear_array(id, a_cookie))
+		return awk_false;
+	freenode((NODE *) a_cookie);
+	return awk_true;
+}
+
 /* api_flatten_array_typed --- flatten out an array so that it can be looped over easily. */
 
 static awk_bool_t
@@ -1523,6 +1583,9 @@ gawk_api_t api_impl = {
 
 	/* Find/open a file */
 	api_get_file,
+
+	/* Additional array hook to destroy an array */
+	api_destroy_array,
 };
 
 /* init_ext_api --- init the extension API */
@@ -1531,12 +1594,13 @@ void
 init_ext_api()
 {
 	/* force values to 1 / 0 */
-	api_impl.do_flags[0] = (do_lint ? 1 : 0);
-	api_impl.do_flags[1] = (do_traditional ? 1 : 0);
-	api_impl.do_flags[2] = (do_profile ? 1 : 0);
-	api_impl.do_flags[3] = (do_sandbox ? 1 : 0);
-	api_impl.do_flags[4] = (do_debug ? 1 : 0);
-	api_impl.do_flags[5] = (do_mpfr ? 1 : 0);
+	api_impl.do_flags[gawk_do_lint] = (do_lint ? 1 : 0);
+	api_impl.do_flags[gawk_do_traditional] = (do_traditional ? 1 : 0);
+	api_impl.do_flags[gawk_do_profile] = (do_profile ? 1 : 0);
+	api_impl.do_flags[gawk_do_sandbox] = (do_sandbox ? 1 : 0);
+	api_impl.do_flags[gawk_do_debug] = (do_debug ? 1 : 0);
+	api_impl.do_flags[gawk_do_mpfr] = (do_mpfr ? 1 : 0);
+	api_impl.do_flags[gawk_do_csv] = (do_csv ? 1 : 0);
 }
 
 /* update_ext_api --- update the variables in the API that can change */
