@@ -2,23 +2,23 @@
  * int_array.c - routines for arrays of integer indices.
  */
 
-/* 
- * Copyright (C) 1986, 1988, 1989, 1991-2013, 2016,
+/*
+ * Copyright (C) 1986, 1988, 1989, 1991-2013, 2016, 2017, 2019, 2020, 2022,
  * the Free Software Foundation, Inc.
- * 
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
- * 
+ *
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
- * 
+ *
  * GAWK is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
@@ -46,10 +46,10 @@ static inline NODE **int_find(NODE *symbol, long k, uint32_t hash1);
 static NODE **int_insert(NODE *symbol, long k, uint32_t hash1);
 static void grow_int_table(NODE *symbol);
 
-afunc_t int_array_func[] = {
+const array_funcs_t int_array_func = {
+	"int",
 	int_array_init,
 	is_integer,
-	null_length,
 	int_lookup,
 	int_exists,
 	int_clear,
@@ -167,49 +167,53 @@ is_integer(NODE *symbol, NODE *subs)
 	 * a[-3]=1; print "-3" in a  -- true
 	 */
 
-	if ((subs->flags & (STRING|STRCUR)) != 0) {
-		char *cp = subs->stptr, *cpend, *ptr;
-		char save;
-		size_t len = subs->stlen;
+	/* must be a STRING */
+	char *cp = subs->stptr, *cpend, *ptr;
+	char save;
+	size_t len = subs->stlen;
 
-		if (len == 0 || (! isdigit((unsigned char) *cp) && *cp != '-'))
-			return NULL;
-		if (len > 1 && 
-			((*cp == '0')		/* "00", "011" .. */
-				|| (*cp == '-' && *(cp + 1) == '0')	/* "-0", "-011" .. */
-			)
+	if (len == 0 || (! isdigit((unsigned char) *cp) && *cp != '-'))
+		return NULL;
+
+	if (len > 1 &&
+		((*cp == '0')		/* "00", "011" .. */
+			|| (*cp == '-' && *(cp + 1) == '0')	/* "-0", "-011" .. */
 		)
-			return NULL;
-		if (len == 1 && *cp != '-') {	/* single digit */
-			subs->numbr = (long) (*cp - '0');
-			if ((subs->flags & MAYBE_NUM) != 0) {
-				subs->flags &= ~MAYBE_NUM;
-				subs->flags |= NUMBER;
-			}
-			subs->flags |= (NUMCUR|NUMINT);
-			return & success_node;
-		}
-
-		cpend = cp + len;
-		save = *cpend;
-		*cpend = '\0';
-
-		errno = 0;
-		l = strtol(cp, & ptr, 10);
-		*cpend = save;
-		if (errno != 0 || ptr != cpend)
-			return NULL;
-		subs->numbr = l;
-		if ((subs->flags & MAYBE_NUM) != 0) {
-			subs->flags &= ~MAYBE_NUM;
+	)
+		return NULL;
+	if (len == 1 && *cp != '-') {	/* single digit */
+		subs->numbr = (long) (*cp - '0');
+		if ((subs->flags & USER_INPUT) != 0) {
+			/* leave USER_INPUT set */
+			subs->flags &= ~STRING;
 			subs->flags |= NUMBER;
 		}
-		subs->flags |= NUMCUR;
-		if (l <= INT32_MAX && l >= INT32_MIN) {
-			subs->flags |= NUMINT;
-			return & success_node;
-		}
+		subs->flags |= (NUMCUR|NUMINT);
+		return & success_node;
 	}
+
+	cpend = cp + len;
+	save = *cpend;
+	*cpend = '\0';
+
+	errno = 0;
+	l = strtol(cp, & ptr, 10);
+	*cpend = save;
+	if (errno != 0 || ptr != cpend)
+		return NULL;
+
+	subs->numbr = l;
+	if ((subs->flags & USER_INPUT) != 0) {
+		/* leave USER_INPUT set */
+		subs->flags &= ~STRING;
+		subs->flags |= NUMBER;
+	}
+	subs->flags |= NUMCUR;
+	if (l <= INT32_MAX && l >= INT32_MIN) {
+		subs->flags |= NUMINT;
+		return & success_node;
+	}
+
 	return NULL;
 #endif /* CHECK_INTEGER_USING_FORCE_NUMBER */
 }
@@ -232,13 +236,13 @@ int_lookup(NODE *symbol, NODE *subs)
 	/*
 	 * N.B: symbol->table_size is the total # of non-integers (symbol->xarray)
 	 *	and integer elements. Also, symbol->xarray must have at least one
-	 *	item in it, and can not exist if there are no integer elements.
+	 *	item in it, and cannot exist if there are no integer elements.
 	 * 	In that case, symbol->xarray is promoted to 'symbol' (See int_remove).
 	 */
 
 
 	if (! is_integer(symbol, subs)) {
-		xn = symbol->xarray; 
+		xn = symbol->xarray;
 		if (xn == NULL) {
 			xn = symbol->xarray = make_array();
 			xn->vname = symbol->vname;	/* shallow copy */
@@ -256,9 +260,9 @@ int_lookup(NODE *symbol, NODE *subs)
  	hash1 = int_hash(k, symbol->array_size);
 	if ((lhs = int_find(symbol, k, hash1)) != NULL)
 		return lhs;
-	
+
 	/* It's not there, install it */
-	
+
 	symbol->table_size++;
 
 	/* first see if we would need to grow the array, before installing */
@@ -272,7 +276,7 @@ int_lookup(NODE *symbol, NODE *subs)
 		/* have to recompute hash value for new size */
 		hash1 = int_hash(k, symbol->array_size);
 	}
-	
+
 	return int_insert(symbol, k, hash1);
 }
 
@@ -326,7 +330,7 @@ int_clear(NODE *symbol, NODE *subs ATTRIBUTE_UNUSED)
 				r = b->aivalue[j];
 				if (r->type == Node_var_array) {
 					assoc_clear(r);	/* recursively clear all sub-arrays */
-					efree(r->vname);			
+					efree(r->vname);
 					freenode(r);
 				} else
 					unref(r);
@@ -452,10 +456,9 @@ int_copy(NODE *symbol, NODE *newsymb)
 
 	/* find the current hash size */
 	cursize = symbol->array_size;
-	
+
 	/* allocate new table */
-	emalloc(new, BUCKET **, cursize * sizeof(BUCKET *), "int_copy");
-	memset(new, '\0', cursize * sizeof(BUCKET *));
+	ezalloc(new, BUCKET **, cursize * sizeof(BUCKET *), "int_copy");
 
 	old = symbol->buckets;
 
@@ -491,7 +494,7 @@ int_copy(NODE *symbol, NODE *newsymb)
 			newchain->ainext = NULL;
 			pnew = & newchain->ainext;
 		}
-	}	
+	}
 
 	if (symbol->xarray != NULL) {
 		NODE *xn, *n;
@@ -537,7 +540,7 @@ int_list(NODE *symbol, NODE *t)
 	if ((assoc_kind & (AINDEX|AVALUE)) == (AINDEX|AVALUE))
 		elem_size = 2;
 	list_size = elem_size * num_elems;
-	
+
 	if (symbol->xarray != NULL) {
 		xn = symbol->xarray;
 		list = xn->alist(xn, t);
@@ -557,7 +560,7 @@ int_list(NODE *symbol, NODE *t)
 				/* index */
 				num = b->ainum[j];
 				if ((assoc_kind & AISTR) != 0) {
-					sprintf(buf, "%ld", num); 
+					sprintf(buf, "%ld", num);
 					subs = make_string(buf, strlen(buf));
 					subs->numbr = num;
 					subs->flags |= (NUMCUR|NUMINT);
@@ -602,7 +605,7 @@ int_kilobytes(NODE *symbol)
 		for (b = symbol->buckets[i]; b != NULL; b = b->ainext)
 			bucket_cnt++;
 	}
-	kb = (((AWKNUM) bucket_cnt) * sizeof (BUCKET) + 
+	kb = (((AWKNUM) bucket_cnt) * sizeof (BUCKET) +
 			((AWKNUM) symbol->array_size) * sizeof (BUCKET *)) / 1024.0;
 
 	if (symbol->xarray != NULL)
@@ -736,7 +739,7 @@ int_hash(uint32_t k, uint32_t hsize)
  */
 
 	/* This is the final mixing function used by Paul Hsieh in SuperFastHash. */
- 
+
 	k ^= k << 3;
 	k += k >> 5;
 	k ^= k << 4;
@@ -778,7 +781,7 @@ int_insert(NODE *symbol, long k, uint32_t hash1)
 
 	b = symbol->buckets[hash1];
 
-	/* Only the first bucket in the chain can be partially full, but is never empty. */ 
+	/* Only the first bucket in the chain can be partially full, but is never empty. */
 
 	if (b == NULL || (i = b->aicount) == 2) {
 		getbucket(b);
@@ -789,7 +792,7 @@ int_insert(NODE *symbol, long k, uint32_t hash1)
 	}
 
 	b->ainum[i] = k;
-	b->aivalue[i] = dupnode(Nnull_string);
+	b->aivalue[i] = new_array_element();
 	b->aicount++;
 	return & b->aivalue[i];
 }
@@ -818,7 +821,7 @@ grow_int_table(NODE *symbol)
 	static const unsigned long sizes[] = {
 		13, 127, 1021, 8191, 16381, 32749, 65497,
 		131101, 262147, 524309, 1048583, 2097169,
-		4194319, 8388617, 16777259, 33554467, 
+		4194319, 8388617, 16777259, 33554467,
 		67108879, 134217757, 268435459, 536870923,
 		1073741827
 	};
@@ -838,8 +841,7 @@ grow_int_table(NODE *symbol)
 	}
 
 	/* allocate new table */
-	emalloc(new, BUCKET **, newsize * sizeof(BUCKET *), "grow_int_table");
-	memset(new, '\0', newsize * sizeof(BUCKET *));
+	ezalloc(new, BUCKET **, newsize * sizeof(BUCKET *), "grow_int_table");
 
 	old = symbol->buckets;
 	symbol->buckets = new;
